import numpy as np
import json
import csv
from typing import NamedTuple
from pathlib import Path
import scipy.stats

TrajectoryPoint = NamedTuple('TrajectoryPoint', x=float, y=float, z=float,
                             t=float, theta=float)


class Trajectory():
    """
    Represents a trajectory.

    Parameters
    ----------
    x : np.ndarray
        Array containing position data of X axis.
    y : np.ndarray
        Array containing position data of Y axis. (Default is None).
    z : np.ndarray
        Array containing position data of X axis. (Default is None).
    t : np.ndarray
        Array containing time data. (Default is None).
    theta : np.ndarray
        Array containing angle data. (Default is None).
    dt : float
        If no time data (``t``) is given this represents the time
        between each position data value.
    id : str
        Id of the trajectory.

    Attributes
    ----------
    x : np.ndarray
        Array containing position data of X axis.
    y : np.ndarray
        Array containing position data of Y axis.
    z : np.ndarray
        Array containing position data of X axis.
    t : np.ndarray
        Array containing time data.
    theta : np.ndarray
        Array containing angle data.
    dt : float
        If no time data (``t``) is given this represents the time
        between each position data value.
    id : str
        Id of the trajectory.

    Raises
    ------
    ValueError
        If ``x`` is not given.
    ValueError
        If all the given position data (``x``, ``y`` and/or ``z``)
        does not have the same shape.
    """

    def __init__(self, x: np.ndarray, y: np.ndarray = None,
                 z: np.ndarray = None, t: np.ndarray = None,
                 theta: np.ndarray = None, dt: float = 1, 
                 id: str = None):

        if x is None:
            raise ValueError('Trajectory requires at least one dimension')
        elif y is not None and z is None:
            if len(x) != len(y):
                raise ValueError('X and Y arrays must have the same shape')
        elif y is not None and z is not None:
            if len(x) != len(y) != len(z):
                raise ValueError('X and Z arrays must have the same shape')
        if t is not None:
            if len(x) != len(t):
                raise ValueError('X and Time arrays must have the same shape')
        if theta is not None:
            if len(x) != len(theta):
                raise ValueError('X and Theta arrays must have the same shape')

        self.x = x
        self.y = y
        self.z = z
        self.t = t
        self.theta = theta

        if self.x is not None:
            self.x = np.array(x)
        if self.y is not None:
            self.y = np.array(y)
        if self.z is not None:
            self.z = np.array(z)
        if self.t is not None:
            self.t = np.array(t)
        if self.theta is not None:
            self.theta = np.array(theta)

        self.dt = dt
        self.id = id
    
    def __len__(self):
        return len(self.x)

    def __iter__(self):
        current_time = 0
        for i in range(len(self)):

            x = self.x[i]
            y = self.y[i]     

            y, z, t, theta, = None, None, None, None

            if self.y is not None:
                y = self.y[i]
            if self.z is not None:
                z = self.z[i]
            if self.t is not None:
                t = self.t[i]
            elif self.dt is not None:
                t = current_time
                current_time += self.dt
            if self.theta is not None:
                theta = self.theta[i]

            yield TrajectoryPoint(x, y, z, t, theta)

    def save(self, file_name: str, path: str = '.', file_type: str = 'json',
               overwrite: bool = True):
        """
        Saves a trajectory

        Parameters
        ----------
        file_name : str
            Name of the file.
        path : str
            Path where to save the trajectory. (Default is ``'.'``).
        file_time : str
            Type of the file. (Default is ``json``).

            The only types avaliable are: ``json`` and ``csv``.
        overwrite : bool
            Wheter or not to overwrite the file if it already exists. (Default
            is True)

        Raises
        ------        
        ValueError
            If ``override`` parameter is ``False`` and the file already exists.
        ValueError
            If ``file_type`` is not ``json`` or ``csv``.
        """

        # Contruct full path
        full_path = Path(path) / Path(f'{file_name}.{file_type}')

        # Check file existance
        if not overwrite and full_path.exists():
            raise ValueError(f"File '{str(full_path)}' already exist")

        def convert_to_list(array_data):
            if array_data is None:
                return array_data
            if array_data is not list:
                array_data = list(array_data)
            return array_data

        if file_type == 'json':
            json_dict = {
                'dt' : self.dt,
                'id' : self.id,
                'x' : convert_to_list(self.x),
                'y' : convert_to_list(self.y),
                'z' : convert_to_list(self.z),
                't' : convert_to_list(self.t),
                'theta' : convert_to_list(self.theta)
            }
            with open(str(full_path), 'w') as f:
                json.dump(json_dict, f)

        elif file_type == 'csv':
            with open(str(full_path), 'w', newline='') as f:
                writer = csv.writer(f, delimiter=',')
                writer.writerow([self.id, self.dt])
                for tp in self:
                    writer.writerow([tp.x, tp.y, tp.z, tp.t, tp.theta])
        else:
            raise ValueError(f"Invalid export file type '{file_type}'")

    @staticmethod
    def load(file_path: str):
        """
        Loads a trajectory

        Parameters
        ----------
        file_path : str
            Path of the trajectory file
        
        Returns
        -------
        Trajecotry
            Trajectory loaded.

        Raises
        ------
        ValueError
            If ``file_path`` is a non existing path.
        ValueError
            If ``file_path`` is a not a file.
        ValueError
            If ``file_path`` extension is not ``json`` or ```csv``.        
        """

        path = Path(file_path)

        # Check valid path
        if not path.exists():
            raise ValueError('Path does not exist.')
        if not path.is_file():
            raise ValueError("Path must be a file.")
    
        # Check valid file type
        file_type = path.suffix
        if not path.suffix in ['.json', '.csv']:
            raise ValueError("Invalid file type.")

        with open(file_path, 'r') as f:
            if file_type == '.json':

                data = json.load(f)
                dt = data['dt']
                traj_id = data['id']
                x = data['x']
                y = data['y']
                z = data['z']
                t = data['t']
                theta = data['theta']
                return Trajectory(x=x, y=y, z=z,
                                  t=t, theta=theta, dt=dt,
                                  id=traj_id)

            elif file_type == '.csv':

                def check_empty_val(val):
                    return None if val == '' else val               

                x, y, z = [], [], []
                t, theta = [], []
                traj_id, dt = None, None

                def add_val(arr, val):
                    if arr is not None:
                        arr.append(val)
                    
                for i, row in enumerate(csv.reader(f)):
                    if i == 0:
                        traj_id = check_empty_val(row[0])
                        dt = check_empty_val(row[1])
                        if dt is not None:
                            dt = float(dt)
                        continue

                    add_val(x, check_empty_val(row[0]))
                    add_val(y, check_empty_val(row[1]))
                    add_val(z, check_empty_val(row[2]))
                    add_val(t, check_empty_val(row[3]))
                    add_val(theta, check_empty_val(row[4]))
                
                x = None if not x else x
                y = None if not y else y
                z = None if not z else z
                t = None if not t else t
                theta = None if not theta else theta

                return Trajectory(x=x, y=y, z=z,
                                  t=t, theta=theta, dt=dt,
                                  id=traj_id)
                                  
    def get_t_diff(self):
        if self.t is not None:
            return np.ediff1d(self.t)

    def get_x_diff(self):
        return np.ediff1d(self.x)

    def get_y_diff(self):
        if self.y is not None:
            return np.ediff1d(self.y)

    def get_z_diff(self):
        if self.z is not None:
            return np.ediff1d(self.z)

    def get_theta_diff(self):
        if self.theta is not None:
            return np.ediff1d(self.theta)

    def get_diff(self):
        dx = self.get_x_diff()
        dy = self.get_y_diff()
        if dy is not None:
            dz = self.get_y_diff()
            if dz is not None:
                return np.sqrt(dx**2 + dy**2 + dz**2)
            else:
                return np.sqrt(dx**2 + dy**2)
        else:
            return dx

    def get_x_velocity(self):
        return self.get_x_diff()/self.dt

    def get_y_velocity(self):
        if self.y is not None:
            return self.get_y_diff()/self.dt

    def get_z_velocity(self):
        if self.z is not None:
            return self.get_z_diff()/self.dt

    def get_theta_velocity(self):
        if self.theta is not None:
            return self.get_theta_diff()/self.dt

    def get_velocity(self):
        return self.get_diff()/self.dt




"""
Analyzing package dosctring
"""

from yupi.analyzing.transformations import (
    add_dynamic_reference,
    subsample_trajectory,
    wrap_theta
)

from yupi.analyzing.statistics import (
    estimate_turning_angles,
    estimate_msd,
    estimate_kurtosis,
    estimate_vacf
)
from yupi.analyzing.visualization import plot_trajectories

__all__ = [
    'add_dynamic_reference',
    'plot_trajectories',
    'subsample_trajectory',
    'wrap_theta',
    'estimate_turning_angles',
    'estimate_msd',
    'estimate_kurtosis',
    'estimate_vacf'
]


import numpy as np
import abc
from yupi import Trajectory

class Generator():

    """docstring for Generator"""

    def __init__(self, T:float, dim:int=1, N:int=1, dt:int=1):
        # siulation parameters
        self.T = T            # total time
        self.dim = dim        # trajectory dimensions
        self.N = N            # number of trajectories
        self.dt = dt          # time step of the simulation
        self.n = int(T / dt)  # number of time steps

    @abc.abstractmethod
    def generate(self):
        pass



class RandomWalkGenerator(Generator):
    """
    Multidimensional Random Walk class.
    """
    
    def __init__(self, T:float, dim:int=1, N:int=1, dt:int=1,
            actions:np.ndarray=None, 
            actions_prob:np.ndarray=None, 
            jump_len:np.ndarray=None):

        super().__init__(T, dim, N, dt) 

        # dynamic variables
        self.t = np.arange(self.n) * dt      # time array
        self.r = np.zeros((self.n, dim, N))  # position array

        # model parameters
        # only right/left jumps, uniform probabilities for it
        # and equal length for all jumps are set as default
        # TODO: Check that the model parameters received have the expected shape
        self.actions = np.array([1, -1]) if actions is None else actions
        self.actions_prob = np.tile([.5, .5], (dim, 1)) if actions_prob is None else actions_prob
        self.jump_len = np.ones((dim, N)) if jump_len is None else jump_len



    # compute vector position as a function of time for
    # all the walkers of the ensemble
    def get_r(self):
        # get movements for every space coordinates according 
        # to the sample space of probabilities in self.actions_prob
        dr = [np.random.choice(self.actions, p=p, size=(self.n - 1, self.N)) for p in self.actions_prob]
        
        # set time/coordinates as the first/second axis
        dr = np.swapaxes(dr, 0, 1)
        
        # scale displacements according to the jump length statistics
        dr = dr * self.jump_len

        # integrate displacements to get position vectors
        self.r[1:] = np.cumsum(dr, axis=0)
        return self.r


    def generate(self):
        # get RandomWalk object and get position vectors
        r = self.get_r()

        trajs  = []
        for i in range(self.N):
            x = r[:,0,i]
            y = r[:,1,i] if self.dim > 1 else None
            z = r[:,2,i] if self.dim > 2 else None
            trajs.append(Trajectory(x=x, y=y, z=z, dt=self.dt, t=self.t,
                                  id="Random Walker {}".format(i+1)))
        return trajs

class LangevinGenerator(Generator):
    """
    Random Walk class from a multidimensional Langevin Equation.
    """

    def __init__(self, T:float, dim:int=1, N:int=1, dt:int=1,
        tau:float=1.,
        noise_pdf:str='normal',
        noise_scale:float=1,
        v0:np.ndarray=None, r0:np.ndarray=None):
        
        super().__init__(T, dim, N, dt) 

        # model parameters
        self.tau = tau                  # relaxation characteristic time
        self.noise_pdf = noise_pdf      # noise PDF
        self.noise_scale = noise_scale  # scale parameter (not stan. dev.)
        self.noise = np.ndarray         # noise array that will be fill in get_noise method

        # dynamic variables
        self.shape = (self.n, dim, N)           # shape of the dynamic variables
        self.t = np.linspace(0, T, num=self.n)  # time array
        self.r = np.empty(self.shape)           # position array
        self.v = np.empty(self.shape)           # velocity array

        # initial conditions
        # TODO: Check that r0 have the rigth shape
        self.r[0] = np.zeros((dim, N)) if r0 is None else r0
        # TODO: Check that v0 have the rigth shape
        self.v[0] = np.zeros((dim, N)) if v0 is None else v0

        self.v_scale = 1
        self.r_scale = 1
        self.t_scale = 1

    def set_scale(self, v_scale=None, r_scale=None, t_scale=None):
        if v_scale:
            self.v_scale = v_scale
        if r_scale:
            self.r_scale = r_scale
        if t_scale:
            self.t_scale = t_scale

    # fill noise array with custom noise properties
    def get_noise(self):
        dist = getattr(np.random, self.noise_pdf)
        self.noise = dist(scale=self.noise_scale, size=self.shape)


    # solve Langevin Equation using the numerical method of Euler-Maruyama
    def solve_rv(self):
        for i in range(self.n - 1):
            # solving for position
            self.r[i + 1] = self.r[i] + \
                            self.v[i] * self.dt

            # solving for velocity
            self.v[i + 1] = self.v[i] + \
                            -np.dot(1 / self.tau, self.v[i]) * self.dt + \
                            self.noise[i] * np.sqrt(self.dt)

    # simulate the process
    def simulate(self):
        self.get_noise()  # create the attribute self.noise
        self.solve_rv()   # solve the Langevin equation

    # Generate yupi Trajectory objects
    def generate(self):
        self.simulate()

        self.r *= self.r_scale 
        self.v *= self.v_scale 
        self.t *= self.t_scale 
        self.dt *= self.t_scale 

        trajs  = []
        for i in range(self.N):
            x = self.r[:,0,i]
            y = self.r[:,1,i] if self.dim > 1 else None
            z = self.r[:,2,i] if self.dim > 2 else None
            trajs.append(Trajectory(x=x, y=y, z=z, dt=self.dt, t=self.t,
                                  id="LangevinSolution {}".format(i+1)))
        return trajs



# testing
if __name__ == '__main__':
    from yupi.analyzing.visualization import plot_trajectories
    
    np.random.seed(0)

    print("Testing RandomWalker")
    
    # set parameter values
    T = 500
    dim = 2
    N = 5
    dt = 1
    actions = [1, 0, -1]
    prob = [[.5, .1, .4],
            [.5, 0, .5]]

    # get RandomWalk object and get position vectors
    rw = RandomWalkGenerator(T, dim, N, dt, actions, prob)
    tr = rw.generate()
    plot_trajectories(tr)


import numpy as np
import scipy
from yupi.analyzing import wrap_theta

# relative and cumulative turning angles
def estimate_turning_angles(traj, accumulate=False, 
                    degrees=False, centered=False):
    dx = traj.get_x_diff()
    dy = traj.get_y_diff()
    theta = np.arctan2(dy, dx)

    if not accumulate:
        theta = np.ediff1d(theta)  # relative turning angles
    else:
        theta -= theta[0]          # cumulative turning angles

    return wrap_theta(theta, degrees, centered)


# mean square displacement
# TODO: Fix this implementation for dim != 2 Traj
def estimate_msd(trajs, time_avg=True, lag=None):
    msd = []
    for traj in trajs:
        # ensemble average
        if not time_avg:
            dx = traj.x - traj.x[0]
            dy = traj.y - traj.y[0]
            dr_2 = dx**2 + dy**2

        # time average
        else:
            dr_2 = np.empty(lag)
            for lag_ in range(1, lag + 1):
                dx = traj.x[lag_:] - traj.x[:-lag_]
                dy = traj.y[lag_:] - traj.y[:-lag_]
                dr_2[lag_ - 1] = np.mean(dx**2 + dy**2)

        # append all squared displacements
        msd.append(dr_2)
    
    msd = np.transpose(msd)
    msd_mean = np.mean(msd, axis=1)
    msd_std = np.std(msd, axis=1)
    return msd_mean, msd_std


# get displacements for ensemble average and
# kurtosis for time average
# TODO: Fix this implementation for dim != 2 Traj
def estimate_kurtosis(trajs, time_avg=True, lag=None):
    kurtosis = []
    for traj in trajs:
        if not time_avg:
            dx = traj.x - traj.x[0]
            dy = traj.y - traj.y[0]
            kurt = np.sqrt(dx**2 + dy**2)

        # time average
        else:
            kurt = np.empty(lag)
            for lag_ in range(1, lag + 1):
                dx = traj.x[lag_:] - traj.x[:-lag_]
                dy = traj.y[lag_:] - traj.y[:-lag_]
                dr = np.sqrt(dx**2 + dy**2)
                kurt[lag_ - 1] = scipy.stats.kurtosis(dr, fisher=False)

        kurtosis.append(kurt)

    if not time_avg:
        return scipy.stats.kurtosis(kurtosis, axis=0, fisher=False)
    else:
        return np.mean(kurtosis, axis=0)


# velocity autocorrelation function
# TODO: Fix this implementation for dim != 2
def estimate_vacf(trajs, time_avg=True, lag=None):
    vacf = []
    for traj in trajs:
        vx = traj.get_x_velocity()
        vy = traj.get_y_velocity()

        # ensemble average
        if not time_avg:
            v1v2x = vx[0] * vx
            v1v2y = vy[0] * vy
            v1v2 = v1v2x + v1v2y

        # time average
        else:
            v1v2 = np.empty(lag)
            for lag_ in range(1, lag + 1):
                v1v2x = vx[:-lag_] * vx[lag_:]
                v1v2y = vy[:-lag_] * vy[lag_:]
                v1v2[lag_ - 1] = np.mean(v1v2x + v1v2y)
        
        # append all pair-wise veloctiy dot products
        vacf.append(v1v2)
    
    vacf = np.transpose(vacf)
    vacf_mean = np.mean(vacf, axis=1)
    vacf_std = np.std(vacf, axis=1)
    return vacf_mean, vacf_std



import matplotlib.pyplot as plt

def plot_trajectories(trajectories):
    # TODO: Check if trajectories is list of Trajectory
    # or Trajectory, if the second case traj = [traj]
    # If none of both case raise exception
    for t in trajectories:

        # plotting
        traj_plot = plt.plot(t.x, t.y, '-')
        color = traj_plot[-1].get_color()
        plt.plot(t.x[0], t.y[0], 'o', mfc='white', zorder=2,
                 label=f'{t.id} initial position', color=color)
        plt.plot(t.x[-1], t.y[-1], 'o', mfc='white', zorder=2,
                 color=color)
        plt.plot(t.x[-1], t.y[-1], 'o', alpha=.5,
                 label=f'{t.id} final position', color=color)

        plt.legend(fontsize=12)

        plt.title('Trajectories in global coordinates')
        plt.tick_params(direction='in', labelsize=12)
        plt.axis('equal')
        plt.grid(True)
        plt.xlabel('x [m]', fontsize=12)
        plt.ylabel('y [m]', fontsize=12)

    plt.show()


"""
Generating package docstring
"""
from yupi.generating.generators import (
    Generator,
    RandomWalkGenerator,
    LangevinGenerator
)

__all__ = [
    'Generator',
    'RandomWalkGenerator',
    'LangevinGenerator'
]